from approvaltests import verify
from approval_utilities.utilities.multiline_string_utils import remove_indentation_from


def test_remove_indentation_from_works_perfectly() -> None:
    text = remove_indentation_from(
        """

        ^^ Blank line above ^^

        Here is some text
          1. with some indentation
          2. and more
            a. even more
          3. little less

        VV Blank line Below VV 

        """
    )
    verify("remove_indentation_from\n" + text)


def test_remove_indentation_uses_the_last_line_as_max_whitespace():
    input = """
        4 whitespaces
    """
    assert "    4 whitespaces\n" == remove_indentation_from(input)


def test_trailing_whitespace():
    input = """
    4 trailing whitespaces    
    """
    assert "4 trailing whitespaces    \n" == remove_indentation_from(input)



from approvaltests import verify_exception
from approval_utilities.utilities.exceptions.exception_collector import (
    ExceptionCollector,
    gather_all_exceptions,
)


def is_odd(integer):
    if integer % 2 != 0:
        raise ValueError(f"{integer} is not odd!")


def test_gather_all_exceptions():
    collector = ExceptionCollector()
    for i in range(1, 6):
        collector.gather(lambda: is_odd(i))

    verify_exception(lambda: collector.release())


def test_gather_all_exceptions_from_list():
    verify_exception(lambda: gather_all_exceptions(range(1, 6), is_odd).release())



import textwrap

from approval_utilities.utilities.logger.simple_logger import SimpleLogger


def remove_indentation_from(text: str) -> str:
    SimpleLogger.variable("text", text)
    if not text:
        return ""
    cleaned = textwrap.dedent(text + "|")
    if cleaned.startswith("\n"):
        cleaned = cleaned[1:]
    return cleaned[:-1]



from abc import abstractmethod
from typing import TypeVar, Generic

T = TypeVar("T")


class Loader(Generic[T]):
    @abstractmethod
    def load(self) -> T:
        raise Exception("Interface member not implemented")



import unittest
from inspect import FrameInfo
from typing import Callable, Any

import pytest

from approval_utilities.utilities.multiline_string_utils import remove_indentation_from
from approvaltests import (
    StackFrameNamer,
    Options,
    verify,
    verify_all,
    verify_all_combinations_with_labeled_input,
    ApprovalException,
)
from approvaltests.inline.inline_options import InlineOptions
from approvaltests.inline.parse_docstring import parse_docstring
from approvaltests.reporters.report_quietly import ReportQuietly
from approvaltests.reporters.report_with_beyond_compare import (
    ReportWithPycharm,
    ReportWithBeyondCompare,
)


def get_approved_via_doc_string():
    test_stack_frame: FrameInfo = StackFrameNamer.get_test_frame()
    method: Callable[..., Any] = get_caller_method(test_stack_frame)
    return remove_indentation_from(method.__doc__)


def get_caller_method(caller_frame) -> Callable:
    caller_function_name: str = caller_frame[3]
    caller_function_object = caller_frame.frame.f_globals.get(
        caller_function_name, None
    )
    return caller_function_object


# Todo:
# detect the actual tab
# detect if the quote type used in the docstring  (i.e. " or ')


def fizz_buzz(param):
    return_string = ""
    for i in range(1, param + 1):
        if i % 15 == 0:
            return_string += "FizzBuzz\n"
        elif i % 3 == 0:
            return_string += "Fizz\n"
        elif i % 5 == 0:
            return_string += "Buzz\n"
        else:
            return_string += str(i) + "\n"
    return return_string


def test_fizz_buzz():
    """
    1
    2
    Fizz
    4
    Buzz
    Fizz
    7
    8
    """
    verify(fizz_buzz(8), options=Options().inline())


def test_docstrings():
    """
    hello
    world
    """
    # verify_inline(greetting())
    # verify(greetting(), options=Options().inline(show_code= False))
    verify(greeting(), options=Options().inline())


def greeting():
    return "hello\nworld"


def test_docstring_parsing():
    """
    1
    2 -> 2
    Fizz
    4
    Buzz
    Fizz
    7
    8
    """
    verify_all("inputs", parse_docstring())


def test_uppercase():
    """
    a -> A
    b -> B
    c -> C
    """
    verify(
        "\n".join([f"{a} -> {a.upper()}" for a in parse_docstring()]),
        options=Options().inline(),
    )


options = Options().inline()


def test_when_options_is_created_outside_of_test():
    """
    hello
    world
    """
    verify(greeting(), options=options)


def test_exception_on_failure():
    """
    this string should not match
    """
    with pytest.raises(ApprovalException):
        verify(greeting(), options=Options().with_reporter(ReportQuietly()).inline())


class InlineTests(unittest.TestCase):
    def test_with_labeled_input_inline(self) -> None:
        """
        (arg1: 1, arg2: 2) => 3
        (arg1: 1, arg2: 4) => 5
        (arg1: 3, arg2: 2) => 5
        (arg1: 3, arg2: 4) => 7
        """
        verify_all_combinations_with_labeled_input(
            lambda a, b: a + b,
            arg1=(1, 3),
            arg2=(2, 4),
            options=Options().inline(),
        )


def get_preceding_whitespace():
    return "    4 whitespaces"


# fmt: off
def test_preceding_whitespace():
    """
        4 whitespaces
    """
    verify(get_preceding_whitespace(), options=Options().inline())


def test_trailing_whitespace():
    """
    4 trailing whitespaces    
    """
    # Note: Pycharm will remove the trailing whitespaces, to disable this go to:
    # File -> Settings -> Editor -> General -> On Save -> [ ] Remove trailing spaces
    verify("4 trailing whitespaces    ", options=Options().inline())

# fmt: on


def test_bug_blank_lines():
    """


    test bug with blank lines



    """
    verify("\n\ntest bug with blank lines\n\n\n\n", options=Options().inline())


def test_inline_with_additional_reporter():
    """
    hello
    world
    """
    verify(
        "hello\nworld",
        options=(Options().inline().add_reporter(ReportWithBeyondCompare())),
    )


def test_inline_with_preserved_approved_text():
    """
    42
    ***** DELETE ME TO APPROVE *****
    vvvvv PREVIOUS RESULT vvvvv
    41
    """
    options = Options().inline(InlineOptions.semi_automatic_with_previous_approved())
    try:
        verify("42", options=options)
    except ApprovalException:
        pass
    verify(get_approved_via_doc_string())


def test_inline_with_semi_automatic_inline():
    """
    42
    ***** DELETE ME TO APPROVE *****
    """
    options = Options().inline(InlineOptions.semi_automatic())
    try:
        verify("42", options=options)
    except ApprovalException:
        pass
    verify(get_approved_via_doc_string())



import unittest

from approvaltests.approval_exception import ApprovalException
from approvaltests.combination_approvals import (
    verify_all_combinations,
    verify_all_combinations_with_namer,
    verify_all_combinations_with_labeled_input,
)
from approvaltests.reporters import CommandLineReporter
from approvaltests.reporters.testing_reporter import ReporterForTesting


class VerifyAllCombinationsTests(unittest.TestCase):
    def setUp(self) -> None:
        self.reporter = None
        self.func = lambda *args: sum(args) + 1

    def test_fails_for_mismatch_with_for_func_accepting_one_arg_and_combination_of_one_arg(
        self,
    ) -> None:
        arg1_combinations = (1,)
        all_args_combinations = (arg1_combinations,)
        with self.assertRaises(ApprovalException):
            verify_all_combinations(
                self.func, all_args_combinations, reporter=ReporterForTesting()
            )

    def test_passes_for_func_accepting_one_arg_and_combination_of_one_arg(self) -> None:
        arg1_combinations = (1,)
        all_args_combinations = (arg1_combinations,)
        verify_all_combinations(
            self.func, all_args_combinations, reporter=self.reporter
        )

    def test_fails_for_mismatch_with_for_func_accepting_one_arg_and_combination_of_two_args(
        self,
    ) -> None:
        arg1_combinations = (1, 2)
        all_args_combinations = (arg1_combinations,)
        with self.assertRaises(ApprovalException):
            verify_all_combinations(
                self.func, all_args_combinations, reporter=ReporterForTesting()
            )

    def test_passes_for_func_accepting_one_arg_and_combination_of_two_args(
        self,
    ) -> None:
        arg1_combinations = (1, 2)
        all_args_combinations = (arg1_combinations,)
        verify_all_combinations(
            self.func, all_args_combinations, reporter=self.reporter
        )

    def test_fails_for_mismatch_with_for_func_accepting_two_args_and_combination_of_one_arg(
        self,
    ) -> None:
        arg1_combinations = (1,)
        arg2_combinations = (2,)
        arg_combinations = (arg1_combinations, arg2_combinations)
        with self.assertRaises(ApprovalException):
            verify_all_combinations(
                self.func, arg_combinations, reporter=ReporterForTesting()
            )

    def test_passes_for_func_accepting_two_args_and_combination_of_one_arg(
        self,
    ) -> None:
        arg1_combinations = (1,)
        arg2_combinations = (2,)
        arg_combinations = (arg1_combinations, arg2_combinations)
        verify_all_combinations(self.func, arg_combinations, reporter=self.reporter)

    def test_fails_for_mismatch_with_for_func_accepting_two_args_and_combination_of_two_args(
        self,
    ) -> None:
        arg1_combinations = (1, 3)
        arg2_combinations = (2, 4)
        arg_combinations = (arg1_combinations, arg2_combinations)
        with self.assertRaises(ApprovalException):
            verify_all_combinations(
                self.func, arg_combinations, reporter=ReporterForTesting()
            )

    def test_for_func_accepting_three_args_and_combination_of_three_args(self) -> None:
        arg1_combinations = (1, 2, 3)
        arg2_combinations = (2, 4, 6)
        arg3_combinations = (10, 11, 12)
        arg_combinations = [arg1_combinations, arg2_combinations, arg3_combinations]
        verify_all_combinations(self.func, arg_combinations, reporter=self.reporter)

    def test_when_arg_combinations_have_equal_lengths(self) -> None:
        arg1_combinations = (1, 3, 5, 7)
        arg2_combinations = (2, 4, 6)
        arg_combinations = (arg1_combinations, arg2_combinations)
        verify_all_combinations(self.func, arg_combinations, reporter=self.reporter)

    def test_records_exception_message_when_function_under_test_throws_an_exception(
        self,
    ) -> None:
        def function_that_raises_exceptions(*args):
            raise Exception(args)

        arg1_combinations = (1, 3)
        arg2_combinations = (2, 4)
        arg_combinations = (arg1_combinations, arg2_combinations)
        verify_all_combinations(
            function_that_raises_exceptions, arg_combinations, reporter=self.reporter
        )

    def test_uses_user_specified_formatter_when_supplied(self) -> None:
        arg1_combinations = (1, 3)
        arg2_combinations = (2, 4)
        arg_combinations = (arg1_combinations, arg2_combinations)
        verify_all_combinations(
            self.func,
            arg_combinations,
            formatter=lambda args, output: "inputs="
            + str(args)
            + ", outputs="
            + str(output)
            + "\n",
            reporter=self.reporter,
        )

    def test_with_labeled_input(self) -> None:
        # begin-snippet: named_combinations
        verify_all_combinations_with_labeled_input(
            self.func,
            arg1=(1, 3),
            arg2=(2, 4),
        )
        # end-snippet


class VerifyAllCombinationsWithNamerTests(unittest.TestCase):
    def setUp(self) -> None:
        self.reporter = CommandLineReporter()
        self.func = lambda *args: sum(args) + 1

    def test_passes_for_func_accepting_one_arg_and_combination_of_one_arg(self) -> None:
        arg1_combinations = (1,)
        all_args_combinations = (arg1_combinations,)
        verify_all_combinations_with_namer(
            self.func, all_args_combinations, reporter=self.reporter
        )


def test_example_for_combinations():
    # begin-snippet: combination_introduction
    verify_all_combinations(is_awake, [["Monday", "Sunday"], ["7:00", "9:00", "11:00"]])
    # end-snippet


def test_starting_snippet():
    # begin-snippet: combinations_starting_point
    inputs1 = ["input1.value1", "input1.value2"]
    inputs2 = ["input2.value1", "input2.value2", "input2.value3"]
    verify_all_combinations(lambda a, b: "placeholder", [inputs1, inputs2])
    # end-snippet


def is_awake(day, time):
    weekdays = ["Monday"]
    is_weekday = day in weekdays
    time = int(time.replace(":00", ""))
    if is_weekday and time > 8:
        return "Yes"
    elif not is_weekday and time < 10:
        return "No"
    else:
        return "Maybe"



def compare_log_and_file_system(log_file_names, fs_file_names):
    return [file for file in fs_file_names if file not in log_file_names]


def test_returns_empty_list_for_two_empty_lists():
    log_file_names = []
    fs_file_names = []
    assert compare_log_and_file_system(log_file_names, fs_file_names) == []


def test_returns_empty_list_for_identical_lists():
    log_file_names = ["a", "b", "c"]
    fs_file_names = ["a", "b", "c"]
    assert compare_log_and_file_system(log_file_names, fs_file_names) == []


def test_returns_items_only_in_fs():
    log_file_names = ["a", "b", "c"]
    fs_file_names = ["a", "b", "c", "d", "e"]
    assert compare_log_and_file_system(log_file_names, fs_file_names) == ["d", "e"]


# Future tests
# 1 - fs:[a,b,c], log:[a,b,c] => []
# 2 - fs:[a,b,c,d], log:[a,b,c] => [d]



import os
import ast


def test_no_imports_from_build_directory():
    # Make sure no file imports from build
    root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
    for dirpath, dirnames, filenames in os.walk(root_dir):
        # Exclude directories that start with '.'
        dirnames[:] = [d for d in dirnames if not d.startswith(".") and d != "build"]
        # also exclude venv
        dirnames[:] = [d for d in dirnames if not d.startswith("venv")]
        for filename in filenames:
            # Skip files that start with '.' and non-Python files
            if filename.startswith(".") or not filename.endswith(".py"):
                continue
            file_path = os.path.join(dirpath, filename)
            with open(file_path, "r", encoding="utf-8") as file:
                source = file.read()
            try:
                tree = ast.parse(source, filename=file_path)
            except SyntaxError:
                continue  # Skip files with syntax errors
            for node in ast.walk(tree):
                if isinstance(node, ast.Import):
                    for alias in node.names:
                        if alias.name == "build" or alias.name.startswith("build."):
                            assert False, f"{file_path} imports 'build'"
                elif isinstance(node, ast.ImportFrom):
                    if node.module == "build" or (
                        node.module and node.module.startswith("build.")
                    ):
                        assert False, f"{file_path} imports from 'build'"



from abc import ABC, abstractmethod

from approval_utilities.approvaltests.core.verify_parameters import VerifyParameters


class Verifiable(ABC):
    @abstractmethod
    def get_verify_parameters(self, options: "Options") -> VerifyParameters:
        pass



from typing import Sequence

from approval_utilities.utilities.exceptions.exception_utils import to_string


class MultipleExceptions(Exception):
    def __init__(self, exceptions: Sequence[Exception]):
        msg = "\n  " + "\n  ".join(map(to_string, exceptions))
        super().__init__(msg)



import threading
from abc import ABC, abstractmethod


class Wrapper(ABC):
    @abstractmethod
    def get(self):
        pass


class SingleWrapper(Wrapper):
    def __init__(self, instance):
        self.instance = instance

    def get(self):
        return self.instance


class ThreadedWrapper(Wrapper):
    def __init__(self, generator):
        self.generator = generator
        self.local = threading.local()

    def get(self):
        if not hasattr(self.local, "value"):
            self.local.value = self.generator()
        return self.local.value



import os
import sys
import tempfile
import subprocess

from approvaltests import Options, verify_argument_parser
from approvaltests.utilities.logger.simple_logger_approvals import verify_simple_logger
from approval_utilities.utilities.logger.simple_logger import SimpleLogger
from tests.find_stale_approved_files import create_argument_parser
from approvaltests.scrubbers import create_regex_scrubber


# Create a temporary sandbox directory and log file
def create_sandbox(approved_files, log_entries, nested=False):
    sandbox_dir = tempfile.TemporaryDirectory()
    log_file_path = os.path.join(sandbox_dir.name, "approvedfiles.log")

    # Create approved files in the sandbox
    for file_name in approved_files:
        if nested:
            nested_dir = os.path.join(sandbox_dir.name, "nested")
            os.makedirs(nested_dir, exist_ok=True)
            open(os.path.join(nested_dir, file_name), "w").close()
        else:
            open(os.path.join(sandbox_dir.name, file_name), "w").close()

    # Write the log entries to the log file
    with open(log_file_path, "w") as log_file:
        for entry in log_entries:
            if nested:
                log_file.write(os.path.join(sandbox_dir.name, "nested", entry) + "\n")
            else:
                log_file.write(os.path.join(sandbox_dir.name, entry) + "\n")

    return sandbox_dir, log_file_path


# Execute the comparison script
def execute_script(directory, log_file):

    script = "find_stale_approved_files.py"
    dirname = os.path.dirname(__file__)
    full_script = os.path.join(dirname, script)
    print(f"Executing script: {full_script=}")
    result = subprocess.run(
        [sys.executable, full_script, directory, log_file],
        capture_output=True,
        text=True,
    )
    output = result.stdout
    SimpleLogger.message(output)


def test_create_argument_parser():
    verify_argument_parser(create_argument_parser())


def test_find_stale_approved_files():
    scrubber = create_regex_scrubber(r".+(?=file\d\.)", "")
    with verify_simple_logger(options=Options().with_scrubber(scrubber)):
        # Test Scenario 1: All approved files are in the log
        SimpleLogger.message("Test Scenario 1: All approved files are in the log")
        approved_files_1 = [
            "file1.approved.txt",
            "file2.approved.doc",
            "file3.approved.csv",
        ]
        log_entries_1 = approved_files_1
        verify_files(approved_files_1, log_entries_1)

        # Test Scenario 2: Approved files that are not in the log
        SimpleLogger.message("Test Scenario 2: Approved files that are not in the log")
        approved_files_2 = [
            "file1.approved.txt",
            "file2.approved.doc",
            "file3.approved.csv",
            "file4.approved.md",
        ]
        log_entries_2 = approved_files_2[
            :-1
        ]  # Exclude 'file4.approved.md' from the log
        verify_files(approved_files_2, log_entries_2)

        # Test Scenario 3: Log contains files that are not in the directory
        SimpleLogger.message(
            "Test Scenario 3: Log contains files that are not in the directory"
        )
        approved_files_3 = ["file1.approved.txt", "file2.approved.doc"]
        log_entries_3 = approved_files_3 + [
            "file3.approved.csv"
        ]  # 'file3.approved.csv' is in the log but not in the directory
        verify_files(approved_files_3, log_entries_3)

        # Test Scenario 4: No approved files in the directory
        SimpleLogger.message("Test Scenario 4: No approved files in the directory")
        verify_files([], ["file1.approved.txt", "file2.approved.doc"])

        # Test Scenario 5: Directory has files but none are in the log
        SimpleLogger.message(
            "Test Scenario 5: Directory has files but none are in the log"
        )
        verify_files(["file1.approved.txt", "file2.approved.doc"], [])

        # Test Scenario 6: Nested folders with approved files
        SimpleLogger.message("Test Scenario 6: Nested folders with approved files")
        approved_files_6 = ["file1.approved.txt", "file2.approved.doc"]
        verify_files(approved_files_6, approved_files_6, nested=True)

        # Test Scenario 7: Files that do not match the naming convention
        SimpleLogger.message(
            "Test Scenario 7: Files that do not match the naming convention"
        )
        approved_files_7 = ["file1.txt", "file2.doc", "file3.csv"]
        verify_files(approved_files_7, approved_files_7)

        SimpleLogger.message(
            "Test Scenario 8: Log contains files that are not in the directory"
        )
        approved_files = ["custom_name_1.txt"]
        log_entries = approved_files + ["custom_name_2.txt"]
        verify_files(approved_files, log_entries)


def verify_files(approved_files, log_entries, nested=False):
    SimpleLogger.variable("Approved Files", approved_files)
    SimpleLogger.variable("Log Entries", log_entries)

    sandbox_dir, log_file_path = create_sandbox(
        approved_files, log_entries, nested=nested
    )
    execute_script(sandbox_dir.name, log_file_path)
    sandbox_dir.cleanup()
